import configparser
import logging
from pathlib import Path


def load_config():
    """Creates config-parser, reading the config and returning the read config

    Returns:
        Parsed config
    """
    config = configparser.ConfigParser()
    config.read(PACKAGE_ROOT_PATH / 'config.ini')
    return config


def setup_logging():
    logger = logging.getLogger('archiTop')
    logger.setLevel(load_config()['APP']['LOG_LEVEL'])
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    # Attach stdout handler to logger (log to console)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def getLogger(name: str):
    return logging.getLogger('archiTop.' + name)


PACKAGE_ROOT_PATH = Path(__file__).parent
setup_logging()



"""Sourcefile containing class to interact with archidekt services"""
from typing import List

from archiTop.base_classes import DeckFetcher
from archiTop.data_types import RawCard


class ArchidektFetcher(DeckFetcher):
    """ArchidektFetcher class, implementing abstract baseclass DeckFetcher"""
    base_url = 'https://archidekt.com/api/decks/%s/small/'

    def _parse_single_card(self, card: dict) -> RawCard:
        """Parses single card information from deck service into Card object.

        Args:
            card:   Card json object to parse information from

        Returns:
            Card class containing parsed information from card json object
        """
        card_data = card['card']

        name = card_data['oracleCard']['name']
        quantity = card['quantity']

        edition_code = card_data['edition']['editioncode']
        commander = card['category'] == 'Commander' or 'Commander' in card.get('categories', ())

        return RawCard(name, quantity, edition_code, commander)

    @staticmethod
    def _parse_card_data(raw_deck_data: dict) -> List[dict]:
        """Parses card information from deck data fetched by `_get_raw_deck_data()`.

        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            List of card json objects contained in deck
        """
        return raw_deck_data['cards']

    @staticmethod
    def _parse_deck_name(raw_deck_data: dict) -> str:
        """Parses deck name from deck data fetched by `_get_raw_deck_data()`.

        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            Name of deck
        """
        return raw_deck_data['name']

    @staticmethod
    def _parse_deck_thumbnail_url(raw_deck_data: dict) -> str:
        """Parses thumbnail url from deck data fetched by `_get_raw_deck_data()`.
        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            Thumbnail url for fetched deck information
        """
        return raw_deck_data['featured']

    @staticmethod
    def _validate_single_card_mainboard(card: dict) -> bool:
        """Validates whether a single card belongs to mainboard.

        Args:
            card:   Card json object contained in fetched deck information

        Returns:
            True when card is contained in mainboard, False otherwise
        """
        blacklist = ('Maybeboard', 'Sideboard')
        category_check = card['category'] not in blacklist
        categories_checks = not any([blacklist_entry in card.get('categories', ())
                                     for blacklist_entry in blacklist])

        return category_check and categories_checks



"""Sourcefile containing functionality to load fetched card information from scryfall"""
import pickle
from functools import lru_cache
from pathlib import Path

from archiTop.scryfall import conf, logger, resources_path
from archiTop.scryfall.scryfall_fetcher import syncronize_scryfall_data


def _get_data_path() -> Path:
    if scryfall_data := list(resources_path.glob(conf['BULK_DATA_FILE_PATTERN'])):
        return scryfall_data[0]


@lru_cache
def _load_scryfall_data(file_path: Path):
    logger.debug('Loading scryfall data')
    data = pickle.load(file_path.open('rb'))
    return data


@lru_cache
def _load_scryfall_name_index(file_path: Path):
    data = _load_scryfall_data(file_path)
    return {entry['name']: entry for entry in data}


@lru_cache
def _load_scryfall_id_index(file_path: Path):
    data = _load_scryfall_data(file_path)
    return {entry['id']: entry for entry in data}


@lru_cache
def _load_scryfall_set_name_index(file_path: Path):
    data = _load_scryfall_data(file_path)
    index = {}
    for entry in data:
        index[entry['set']] = index.get(entry['set'], {})
        index[entry['set']][entry['name']] = entry
    return index


def load_scryfall_name_index():
    file_path = _get_data_path()
    return _load_scryfall_name_index(file_path)


def load_scryfall_id_index():
    file_path = _get_data_path()
    return _load_scryfall_id_index(file_path)


def load_scryfall_set_name_index():
    file_path = _get_data_path()
    return _load_scryfall_set_name_index(file_path)



"""Sourcefile containing deck builder class for decks with multiple cards"""
from collections import OrderedDict
from copy import deepcopy

from archiTop.base_classes import DeckBuilder
from archiTop.resources import (card_asset_template, card_deck_template,
                                card_template)
from archiTop.scryfall.data_types import ScryfallCard


class MultiCardDeckBuilder(DeckBuilder):
    """"MultiCardDeckBuilder class implementing abstract DeckBuilder class.
    Used for card decks with multiple cards."""

    def __init__(self, *args):
        self.card_deck_json = deepcopy(card_deck_template)
        self.contained_objects = []
        self.deck_ids = []
        self.custom_deck = OrderedDict()
        super().__init__(*args)

    def __repr__(self):
        unique_cards = len(set(self.deck_ids))
        return f'CardDeck({len(self.deck_ids)} total cards, {unique_cards} unique cards)'

    def _populate_card_template(self, card: ScryfallCard):
        """Creates a new TableTop card object and fills information from card class.
        Each card in deck needs one card object, therefore cards with quantity > 1 will be
        duplicated.
        Same cards, even when duplicated will keep the same ID.

        Once populated, card object is inserted into contained_objects and id added to deck_ids.

        Args:
            card:   Card to create card object for
        """
        card_json = deepcopy(card_template)

        card_json['CardID'] = self.current_card_id
        card_json['Nickname'] = card.tabletop_name

        # create one object per quantity
        for _ in range(card.quantity):
            self.contained_objects.append(card_json)
            self.deck_ids.append(self.current_card_id)

        self.current_card_id += 100

    def _populate_card_asset_template(self, card: ScryfallCard):
        """Creates a new TableTop card asset object and fills with information from card class.
        There should only exist on card asset template for each unique card.
        Therefor cards with quantity > 1 do only get one card asset.

        Asset matching is done with insertion order of asset objects.
        Order in the ContainedObjects, DeckID's must match the order of card assets.

        Once populated, card asset is inserted in custom deck and asset id is incremented.

        Args:
            card:   Card to create asset for
        """
        card_asset_json = deepcopy(card_asset_template)

        card_asset_json['FaceURL'] = card.image_url
        card_asset_json['BackURL'] = self.card_back_url

        self.custom_deck[str(self.current_card_asset_id)] = card_asset_json

        self.current_card_asset_id += 1

    def create_deck(self) -> dict:
        """Create the json structure for the card deck containing multiple cards.

        Returns:
            TableTop card deck json containing multiple cards
        """
        for card in self.cards:
            self._populate_card_template(card)
            self._populate_card_asset_template(card)

        self.card_deck_json['ContainedObjects'] = self.contained_objects
        self.card_deck_json['DeckIDs'] = self.deck_ids
        self.card_deck_json['CustomDeck'] = self.custom_deck
        self.card_deck_json['Transform']['rotZ'] = 180 if self.hidden else 0

        return self.card_deck_json



"""Sourcefile containing functionality to construct scryfall card deck"""
from functools import reduce
from typing import List, Tuple

from archiTop.data_types import RawCard, RawDeck
from archiTop.scryfall.data_types import ScryfallCard, ScryfallDeck
from archiTop.scryfall.scryfall_loader import (load_scryfall_name_index,
                                               load_scryfall_set_name_index,
                                               syncronize_scryfall_data)


class ScryfallDeckBuilder:

    def __init__(self, raw_deck: RawDeck):
        syncronize_scryfall_data()

        # load index to search by name and scryfall id
        self.name_index = load_scryfall_name_index()
        self.set_name_index = load_scryfall_set_name_index()
        self.raw_deck = raw_deck

    def construct_deck(self) -> ScryfallDeck:
        cards, tokens = self._get_scryfall_cards_for_deck()

        return ScryfallDeck(cards, tokens, self.raw_deck.name, self.raw_deck.thumbnail)

    def _get_scryfall_cards_for_deck(self) -> Tuple[List[ScryfallCard], List[ScryfallCard]]:
        mainboard = self.raw_deck.mainboard

        # convert raw cards to scryfall enriched cards
        scryfall_cards = [self._create_scryfall_card(card) for card in mainboard]

        # extract related cards for scryfall cards (token, etc.)
        related_cards = reduce(lambda set1, set2: set1 | set2,
                               [card.related_cards for card in scryfall_cards])
        return scryfall_cards, related_cards

    def _create_scryfall_card(self, card: RawCard) -> ScryfallCard:
        if card.editioncode is not None:
            scryfall_data = self.set_name_index[card.editioncode][card.name]
        else:
            scryfall_data = self.name_index[card.name]
        return ScryfallCard(scryfall_data,
                            quantity=card.quantity,
                            commander=card.commander)



"""Sourcefile containing functionality to fetch card information from scryfall"""
import pickle

import requests

from archiTop.scryfall import conf, logger, resources_path


def _extract_file_name(url: str) -> str:
    file_name = url.split('/')[-1]
    return file_name.replace('.json', '.pkl')


def get_bulk_url() -> str:
    request = requests.get(conf['BULK_META_URL']).json()
    oracle_cards = list(filter(lambda x: x['type'] == conf['BULK_DATA_IDENTIFIER'],
                               request['data']))[0]

    return oracle_cards['download_uri']


def update_scryfall_data(url: str):
    data = requests.get(url).json()
    pickle.dump(data, open(resources_path / _extract_file_name(url), 'wb'))


def syncronize_scryfall_data():
    current_url = get_bulk_url()
    file_name = _extract_file_name(current_url)
    if not (resources_path / file_name).exists():
        logger.info('Re-syncing scryfall data...')
        # delete outdated scryfall data
        for path in resources_path.glob(conf['BULK_DATA_FILE_PATTERN']):
            path.unlink()

        update_scryfall_data(current_url)

        logger.info('Scryfall data is up to date 🎉')



"""Sourcefile containing class fetching deck information"""
from abc import ABC, abstractmethod
from functools import reduce
from typing import List

import requests

from archiTop.config import getLogger
from archiTop.data_types import RawCard, RawDeck

logger = getLogger(__name__)


class DeckFetcher(ABC):
    """Abstract baseclass to for deck fetcher"""
    base_url = None
    mainboard_cards = []

    def __init__(self, deck_id: int):
        """Initializes deck fetcher with id of deck.

        Args:
            deck_id:    DeckID to fetch deck information for
        """
        self.deck_id = deck_id

    def __repr__(self) -> str:
        total_count = reduce(lambda a, b: a + b.quantity, self.mainboard_cards, 0)
        return f'DeckFetcher({total_count} total cards, {len(self.mainboard_cards)} unique cards)'

    def _get_raw_deck_data(self) -> dict:
        """Fetch the deck information by querying base_url combined with the deckID.

        Returns:
            Deck information in json format
        """
        logger.debug('Downloading deck <%s>', self.deck_id)
        return requests.get(self.base_url % self.deck_id).json()

    def get_deck(self) -> RawDeck:
        """Fetches and parses deck information.

        Returns:
            Deck of cards, containing deck information fetched
        """
        raw_deck_data = self._get_raw_deck_data()

        deck_name = self._parse_deck_name(raw_deck_data)
        thumbnail_url = self._parse_deck_thumbnail_url(raw_deck_data)

        thumbnail = requests.get(thumbnail_url).content

        filtered_mainboard_card_data = list(filter(self._validate_single_card_mainboard,
                                                   self._parse_card_data(raw_deck_data)))

        self.mainboard_cards = [self._parse_single_card(card) for card in
                                filtered_mainboard_card_data]

        return RawDeck(self.mainboard_cards, deck_name, thumbnail)

    @abstractmethod
    def _parse_single_card(self, card: dict) -> RawCard:
        """Abstractmethod to be implemented by child class.
        Parses single card information from deck service into Card object.

        Args:
            card:   Card json object to parse information from

        Returns:
            Card class containing parsed information from card json object
        """
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def _parse_card_data(raw_deck_data: dict) -> List[dict]:
        """Abstractmethod to be implemented by child class.
        Parses card information from deck data fetched by `_get_raw_deck_data()`.

        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            List of card json objects contained in deck
        """
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def _parse_deck_name(raw_deck_data: dict) -> str:
        """Abstractmethod to be implemented by child class.
        Parses deck name from deck data fetched by `_get_raw_deck_data()`.

        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            Name of deck
        """
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def _parse_deck_thumbnail_url(raw_deck_data: dict) -> str:
        """Abstractmethod to be implemented by child class.
        Parses thumbnail url from deck data fetched by `_get_raw_deck_data()`.
        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            Thumbnail url for fetched deck information
        """
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def _validate_single_card_mainboard(card: dict) -> bool:
        """Abstractmethod to be implemented by child class.
        Validates whether a single card belongs to mainboard.

        Args:
            card:   Card json object contained in fetched deck information

        Returns:
            True when card is contained in mainboard, False otherwise
        """
        raise NotImplemented



from pathlib import Path

from archiTop.config import PACKAGE_ROOT_PATH, getLogger, load_config

logger = getLogger(__name__)
conf = load_config()['SCRYFALL']
resources_path = Path(PACKAGE_ROOT_PATH, 'resources')

from .scryfall_builder import ScryfallDeckBuilder
from .scryfall_fetcher import syncronize_scryfall_data, syncronize_scryfall_data
from .scryfall_loader import load_scryfall_set_name_index



"""Sourcefile containing final deck builder class for TableTop"""
import json
import sys
from copy import deepcopy
from pathlib import Path
from typing import List, Optional

from archiTop.config import getLogger, load_config
from archiTop.deck_builder.multi_card_deck_builder import MultiCardDeckBuilder
from archiTop.deck_builder.single_card_deck_builder import SingleCardDeckBuilder
from archiTop.resources import final_deck_template
from archiTop.scryfall.data_types import ScryfallCard, ScryfallDeck

logger = getLogger(__name__)


class DeckBuilderWrapper:
    """Wrapper class, converting Deck object into TableTop deck asset"""
    custom_back_url = None

    def __init__(self,
                 deck: ScryfallDeck,
                 custom_back_url: str = None):
        """Initializes deck builder wrapper with deck of cards.

        Args:
            deck:               Deck including cards contained and additional deck information
            custom_back_url:    Url for image to replace the default mtg card-back
        """
        self.final_deck_json = deepcopy(final_deck_template)
        self.card_decks = []

        self.deck = deck
        if custom_back_url:
            logger.debug('Using custom card-back <%s>', custom_back_url)
            self.custom_back_url = custom_back_url

    def construct_final_deck(self):
        """Constructs the final asset json for TableTop.
        Separates mainboard cards from commander, then creates multiple card decks
        for mainboard- and commander-cards.
        """
        mainboard_cards = list(filter(lambda card: not card.commander, self.deck.mainboard))
        commander_cards = list(filter(lambda card: card.commander, self.deck.mainboard))

        # construct related cards card-deck, including tokens
        if token_deck := self._construct_card_deck(self.deck.related_cards, hidden=False):
            self.card_decks.append(token_deck)

        # construct mainboard card-deck
        if mainboard_deck := self._construct_card_deck(mainboard_cards):
            self.card_decks.append(mainboard_deck)

        # construct separate card-deck for commander
        if commander_deck := self._construct_card_deck(commander_cards, hidden=False):
            self.card_decks.append(commander_deck)

        # todo: implement sideboard
        # construct sideboard
        if sideboard_deck := None:
            self.card_decks.append(sideboard_deck)

        # shift deck positions so piles do not combine
        current_x_pos = 0
        for deck in self.card_decks:
            deck['Transform']['posX'] = current_x_pos
            current_x_pos += 2.2

        self.final_deck_json['ObjectStates'] = self.card_decks

    def save_deck(self, export_location: str = None):
        """Saves deck and thumbnail.
        Filename is determined by the chosen deck name.

        If system is mac os, output will be saved to the tabletop location,
        otherwise it will be stored to current directory.

        Args:
            export_location:    Optional argument to overwrite output location
        """
        deck_name = f'{self.deck.name}.json'
        thumbnail_name = f'{self.deck.name}.png'

        if export_location:
            logger.debug(f'Saving deck <{self.deck.name}> to passed location {export_location}')

        else:
            if sys.platform == 'darwin':  # client is using mac os
                logger.debug(f'Saving deck <{self.deck.name}> to tabletop location')
                table_top_save_location = load_config()['EXPORT']['MAC']
                export_location = Path(Path.home(), table_top_save_location)

            else:
                logger.debug(f'Saving deck <{self.deck.name}> to current directory')
                export_location = ''

        # save deck json
        json.dump(self.final_deck_json, open(Path(export_location, deck_name), 'w'))
        # save deck thumbnail
        with open(Path(export_location, thumbnail_name), 'wb') as file:
            file.write(self.deck.thumbnail)

    def _construct_card_deck(self, card_list: List[ScryfallCard],
                             hidden=True) -> Optional[dict]:
        """Chooses DeckBuilder based on amount of amount of cards passed.

        Chooses SingleCardDeckBuilder when card list contains a single card.
        MultiCardDeckBuilder when card list contains multiple cards.

        Args:
            card_list:  List of cards to create deck asset for
            hidden:     Whether deck is hidden or not (face down)
                        Defaults to true.

        Returns:
            Result from deck builder or None when List is empty
        """
        if len(card_list) > 1:
            builder = MultiCardDeckBuilder(card_list, hidden, self.custom_back_url)
        elif len(card_list) == 1:
            builder = SingleCardDeckBuilder(card_list, hidden, self.custom_back_url)
        else:
            logger.warning('Passed card list is empty')
            return None

        return builder.create_deck()
